"use strict";
/*!
 * Copyright (c) Microsoft Corporation and contributors. All rights reserved.
 * Licensed under the MIT License.
 */
Object.defineProperty(exports, "__esModule", { value: true });
exports.createMongoErrorRetryRuleset = void 0;
const IMongoExceptionRetryRule_1 = require("../IMongoExceptionRetryRule");
class InternalErrorRule extends IMongoExceptionRetryRule_1.BaseMongoExceptionRetryRule {
    constructor(retryRuleOverride) {
        super("InternalErrorRule", retryRuleOverride);
        this.defaultRetryDecision = false;
    }
    match(error) {
        return error.code === 1
            && error.codeName
            && error.codeName === InternalErrorRule.codeName;
    }
}
InternalErrorRule.codeName = "InternalError";
// This handles the requested queued on client side buffer overflow. Should relies on reconnect instead of retry?
class NoConnectionAvailableRule extends IMongoExceptionRetryRule_1.BaseMongoExceptionRetryRule {
    constructor(retryRuleOverride) {
        super("NoConnectionAvailableRule", retryRuleOverride);
        this.defaultRetryDecision = false;
    }
    match(error) {
        // TODO: This timed out actually included two different messages:
        // 1. Retries due to rate limiting: False.
        // 2. Retries due to rate limiting: True.
        // We might need to split this into two different rules after consult with DB team.
        return error.message
            && error.message.startsWith(NoConnectionAvailableRule.messagePrefix);
    }
}
NoConnectionAvailableRule.messagePrefix = "no connection available for operation and number of stored operation";
// This handles the no primary found in replicaset or invalid replica set name from client
// Should not retry but relays on reconnect.
class NoPrimaryInReplicasetRule extends IMongoExceptionRetryRule_1.BaseMongoExceptionRetryRule {
    constructor(retryRuleOverride) {
        super("NoPrimaryInReplicasetRule", retryRuleOverride);
        this.defaultRetryDecision = false;
    }
    match(error) {
        // TODO: This timed out actually included two different messages:
        // 1. Retries due to rate limiting: False.
        // 2. Retries due to rate limiting: True.
        // We might need to split this into two different rules after consult with DB team.
        return error.message
            && error.message === NoPrimaryInReplicasetRule.message;
    }
}
NoPrimaryInReplicasetRule.message = "no primary found in replicaset or invalid replica set name";
// this handles the pool destroyed error from client side. Should relies on reconnect instead of retry?
class PoolDestroyedRule extends IMongoExceptionRetryRule_1.BaseMongoExceptionRetryRule {
    constructor(retryRuleOverride) {
        super("PoolDestroyedRule", retryRuleOverride);
        this.defaultRetryDecision = false;
    }
    match(error) {
        return error.code === 16
            && error.message
            && (error.message === PoolDestroyedRule.message1
                || error.message === PoolDestroyedRule.message2);
    }
}
PoolDestroyedRule.message1 = "pool destroyed";
PoolDestroyedRule.message2 = "server instance pool was destroyed";
// this handles the case where the request payload is too large from server.
class RequestSizeLargeRule extends IMongoExceptionRetryRule_1.BaseMongoExceptionRetryRule {
    constructor(retryRuleOverride) {
        super("RequestSizeLargeRule", retryRuleOverride);
        this.defaultRetryDecision = false;
    }
    match(error) {
        return error.code === 16
            && error.errmsg
            && error.errmsg.startsWith(RequestSizeLargeRule.errorMsgPrefix);
    }
}
RequestSizeLargeRule.errorMsgPrefix = "Error=16, Details='Response status code does not indicate success: RequestEntityTooLarge (413)";
// This handles request timeout from server without additional rate limit info.
class RequestTimedNoRateLimitInfo extends IMongoExceptionRetryRule_1.BaseMongoExceptionRetryRule {
    constructor(retryRuleOverride) {
        super("RequestTimedNoRateLimitInfo", retryRuleOverride);
        this.defaultRetryDecision = true;
    }
    match(error) {
        return error.code === 50
            && error.errmsg
            && error.errmsg === RequestTimedNoRateLimitInfo.errmsg;
    }
}
RequestTimedNoRateLimitInfo.errmsg = "Request timed out.";
// This handles request timeout from server with http info.
class RequestTimedOutWithHttpInfo extends IMongoExceptionRetryRule_1.BaseMongoExceptionRetryRule {
    constructor(retryRuleOverride) {
        super("RequestTimedOutWithHttpInfo", retryRuleOverride);
        this.defaultRetryDecision = true;
    }
    match(error) {
        return error.code === 50
            && error.errmsg
            && error.errmsg.startsWith(RequestTimedOutWithHttpInfo.errmsgPrefix);
    }
}
RequestTimedOutWithHttpInfo.errmsgPrefix = "Error=50, Details='Response status code does not indicate success: RequestTimeout (408);";
// This handles request timeout from server with additional rate limit info.
class RequestTimedOutWithRateLimitTrue extends IMongoExceptionRetryRule_1.BaseMongoExceptionRetryRule {
    constructor(retryRuleOverride) {
        super("RequestTimedOutWithRateLimitTrue", retryRuleOverride);
        this.defaultRetryDecision = false;
    }
    match(error) {
        return error.code === 50
            && error.errmsg
            && error.codeName === RequestTimedOutWithRateLimitTrue.codeName
            && error.errmsg === RequestTimedOutWithRateLimitTrue.errorMsg;
    }
}
RequestTimedOutWithRateLimitTrue.codeName = "ExceededTimeLimit";
RequestTimedOutWithRateLimitTrue.errorMsg = "Request timed out. Retries due to rate limiting: True.";
// This handles request timeout from server with additional rate limit info.
class RequestTimedOutWithRateLimitFalse extends IMongoExceptionRetryRule_1.BaseMongoExceptionRetryRule {
    constructor(retryRuleOverride) {
        super("RequestTimedOutWithRateLimitFalse", retryRuleOverride);
        this.defaultRetryDecision = true;
    }
    match(error) {
        return error.code === 50
            && error.errmsg
            && error.codeName === RequestTimedOutWithRateLimitFalse.codeName
            && error.errmsg === RequestTimedOutWithRateLimitFalse.errorMsg;
    }
}
RequestTimedOutWithRateLimitFalse.codeName = "ExceededTimeLimit";
RequestTimedOutWithRateLimitFalse.errorMsg = "Request timed out. Retries due to rate limiting: False.";
// This handles server side temporary 503 issue
class ServiceUnavailableRule extends IMongoExceptionRetryRule_1.BaseMongoExceptionRetryRule {
    constructor(retryRuleOverride) {
        super("ServiceUnavailableRule", retryRuleOverride);
        this.defaultRetryDecision = true;
    }
    match(error) {
        return error.code === 1
            && error.errorDetails
            && error.errorDetails.includes(ServiceUnavailableRule.errorDetails);
    }
}
ServiceUnavailableRule.errorDetails = "Response status code does not indicate success: ServiceUnavailable (503)";
// this handles the pool destroyed error from client side. Should relies on reconnect instead of retry?
class TopologyDestroyed extends IMongoExceptionRetryRule_1.BaseMongoExceptionRetryRule {
    constructor(retryRuleOverride) {
        super("TopologyDestroyed", retryRuleOverride);
        this.defaultRetryDecision = false;
    }
    match(error) {
        return error.message
            && error.message.toLowerCase() === TopologyDestroyed.message;
    }
}
// We see messages with both "Topology was destroyed" and "topology was destroyed" on prod. So need to handle both cases.
TopologyDestroyed.message = "topology was destroyed";
// this handles the incorrect credentials set. Should not retry
class UnauthorizedRule extends IMongoExceptionRetryRule_1.BaseMongoExceptionRetryRule {
    constructor(retryRuleOverride) {
        super("UnauthorizedRule", retryRuleOverride);
        this.defaultRetryDecision = false;
    }
    match(error) {
        return error.code === 13
            && error.codeName
            && error.codeName === UnauthorizedRule.codeName;
    }
}
UnauthorizedRule.codeName = "Unauthorized";
// Maintain the list from more strick faster comparison to less strict slower comparison
function createMongoErrorRetryRuleset(retryRuleOverride) {
    const mongoErrorRetryRuleset = [
        // The rules are using exactly equal
        new InternalErrorRule(retryRuleOverride),
        new NoPrimaryInReplicasetRule(retryRuleOverride),
        new RequestTimedNoRateLimitInfo(retryRuleOverride),
        new RequestTimedOutWithRateLimitTrue(retryRuleOverride),
        new RequestTimedOutWithRateLimitFalse(retryRuleOverride),
        new TopologyDestroyed(retryRuleOverride),
        new UnauthorizedRule(retryRuleOverride),
        // The rules are using multiple compare
        new PoolDestroyedRule(retryRuleOverride),
        // The rules are using string startWith
        new NoConnectionAvailableRule(retryRuleOverride),
        new RequestSizeLargeRule(retryRuleOverride),
        new RequestTimedOutWithHttpInfo(retryRuleOverride),
        // The rules are using string contains
        new ServiceUnavailableRule(retryRuleOverride),
    ];
    return mongoErrorRetryRuleset;
}
exports.createMongoErrorRetryRuleset = createMongoErrorRetryRuleset;
//# sourceMappingURL=index.js.map