"use strict";
/*!
 * Copyright (c) Microsoft Corporation and contributors. All rights reserved.
 * Licensed under the MIT License.
 */
Object.defineProperty(exports, "__esModule", { value: true });
exports.MongoErrorRetryAnalyzer = void 0;
const server_services_telemetry_1 = require("@fluidframework/server-services-telemetry");
const defaultExceptionRule_1 = require("./defaultExceptionRule");
const mongoError_1 = require("./mongoError");
const mongoNetworkError_1 = require("./mongoNetworkError");
class MongoErrorRetryAnalyzer {
    constructor(retryRuleOverride) {
        this.mongoNetworkErrorRetryRuleset = (0, mongoNetworkError_1.createMongoNetworkErrorRetryRuleset)(retryRuleOverride);
        this.mongoErrorRetryRuleset = (0, mongoError_1.createMongoErrorRetryRuleset)(retryRuleOverride);
        this.defaultRule = new defaultExceptionRule_1.DefaultExceptionRule(retryRuleOverride);
    }
    static getInstance(retryRuleOverride) {
        if (!this.instance) {
            this.instance = new MongoErrorRetryAnalyzer(retryRuleOverride);
        }
        return this.instance;
    }
    shouldRetry(error) {
        const rule = this.getRetryRule(error);
        if (!rule) {
            // This should not happen.
            server_services_telemetry_1.Lumberjack.error("MongoErrorRetryAnalyzer.shouldRetry() didn't get a rule", undefined, error);
            return false;
        }
        const ruleName = rule.ruleName;
        const decision = rule.shouldRetry();
        const properties = {
            ruleName,
            decision,
        };
        server_services_telemetry_1.Lumberjack.warning(`Error rule used ${rule.ruleName}, shouldRetry: ${decision}`, properties, error);
        return decision;
    }
    getRetryRule(error) {
        if (error.name === "MongoNetworkError") {
            return this.getRetryRuleFromSet(error, this.mongoNetworkErrorRetryRuleset);
        }
        if (error.name === "MongoError") {
            return this.getRetryRuleFromSet(error, this.mongoErrorRetryRuleset);
        }
        return this.defaultRule;
    }
    getRetryRuleFromSet(error, ruleSet) {
        return ruleSet.find((rule) => rule.match(error)) || this.defaultRule;
    }
}
exports.MongoErrorRetryAnalyzer = MongoErrorRetryAnalyzer;
;
//# sourceMappingURL=mongoErrorRetryAnalyzer.js.map