"use strict";
/*!
 * Copyright (c) Microsoft Corporation and contributors. All rights reserved.
 * Licensed under the MIT License.
 */
Object.defineProperty(exports, "__esModule", { value: true });
exports.MongoDbFactory = exports.MongoDb = exports.MongoCollection = void 0;
const console_1 = require("console");
const mongodb_1 = require("mongodb");
const server_services_telemetry_1 = require("@fluidframework/server-services-telemetry");
const server_services_core_1 = require("@fluidframework/server-services-core");
const mongoExceptionRetryRules_1 = require("./mongoExceptionRetryRules");
const MaxFetchSize = 2000;
const MaxRetryAttempts = 3;
const InitialRetryIntervalInMs = 1000;
class MongoCollection {
    constructor(collection, retryEnabled = false, telemetryEnabled = false, mongoErrorRetryAnalyzer) {
        this.collection = collection;
        this.retryEnabled = retryEnabled;
        this.telemetryEnabled = telemetryEnabled;
        this.mongoErrorRetryAnalyzer = mongoErrorRetryAnalyzer;
    }
    async aggregate(pipeline, options) {
        const req = async () => new Promise((resolve) => resolve(this.collection.aggregate(pipeline, options)));
        return this.requestWithRetry(req, "MongoCollection.aggregate");
    }
    async find(query, sort, limit = MaxFetchSize, skip) {
        const req = async () => {
            let queryCursor = this.collection
                .find(query)
                .sort(sort)
                .limit(limit);
            if (skip) {
                queryCursor = queryCursor.skip(skip);
            }
            return queryCursor.toArray();
        };
        return this.requestWithRetry(req, "MongoCollection.find", query);
    }
    async findOne(query) {
        const req = async () => this.collection.findOne(query);
        return this.requestWithRetry(req, // request
        "MongoCollection.findOne", // callerName
        query);
    }
    async findAll() {
        const req = async () => this.collection.find({}).toArray();
        return this.requestWithRetry(req, // request
        "MongoCollection.findAll");
    }
    async update(filter, set, addToSet) {
        const req = async () => this.updateCore(filter, set, addToSet, false);
        return this.requestWithRetry(req, // request
        "MongoCollection.update", // callerName
        filter);
    }
    async updateMany(filter, set, addToSet) {
        const req = async () => this.updateManyCore(filter, set, addToSet, false);
        return this.requestWithRetry(req, // request
        "MongoCollection.updateMany", // callerName
        filter);
    }
    async upsert(filter, set, addToSet) {
        const req = async () => this.updateCore(filter, set, addToSet, true);
        return this.requestWithRetry(req, // request
        "MongoCollection.upsert", // callerName
        filter);
    }
    async distinct(key, query) {
        const req = async () => this.collection.distinct(key, query);
        return this.requestWithRetry(req, // request
        "MongoCollection.distinct", // callerName
        query);
    }
    async deleteOne(filter) {
        const req = async () => this.collection.deleteOne(filter);
        return this.requestWithRetry(req, // request
        "MongoCollection.deleteOne", // callerName
        filter);
    }
    async deleteMany(filter) {
        const req = async () => this.collection.deleteMany(filter);
        return this.requestWithRetry(req, // request
        "MongoCollection.deleteMany", // callerName
        filter);
    }
    async insertOne(value) {
        const req = async () => {
            const result = await this.collection.insertOne(value);
            return result.insertedId;
        };
        return this.requestWithRetry(req, // request
        "MongoCollection.insertOne", // callerName
        value);
    }
    async insertMany(values, ordered) {
        const req = async () => this.collection.insertMany(values, { ordered: false });
        await this.requestWithRetry(req, // request
        "MongoCollection.insertMany");
    }
    // Create indexes with unique restriction were always failing due to :
    // 1. We have existing data
    // 2. The unique restriction not on partition key for some collections.
    // The error will continue without a DB rebuild, which is a hard and long work.
    // Also create index mostly happened at service bootstrap time. If we bubble up at that time,
    // service will failed to start. So instead we need catch the exception and log it without bubbling up.
    async createIndex(index, unique) {
        const req = async () => this.collection.createIndex(index, { unique });
        try {
            const indexName = await this.requestWithRetry(req, // request
            "MongoCollection.createIndex");
            server_services_telemetry_1.Lumberjack.info(`Created index ${indexName}`);
        }
        catch (error) {
            server_services_telemetry_1.Lumberjack.error(`Index creation failed`, undefined, error);
        }
    }
    // Create index mostly happened at service bootstrap time. If we bubble up at that time,
    // service will failed to start. So instead we need catch the exception and log it without bubbling up.
    async createTTLIndex(index, expireAfterSeconds) {
        const req = async () => this.collection.createIndex(index, { expireAfterSeconds });
        try {
            const indexName = await this.requestWithRetry(req, // request
            "MongoCollection.createTTLIndex");
            server_services_telemetry_1.Lumberjack.info(`Created TTL index ${indexName}`);
        }
        catch (error) {
            server_services_telemetry_1.Lumberjack.error(`TTL Index creation failed`, undefined, error);
        }
    }
    async findOrCreate(query, value) {
        const req = async () => {
            const result = await this.collection.findOneAndUpdate(query, {
                $setOnInsert: value,
            }, {
                returnOriginal: true,
                upsert: true,
            });
            return result.value
                ? { value: result.value, existing: true }
                : { value, existing: false };
        };
        return this.requestWithRetry(req, // request
        "MongoCollection.findOrCreate", // callerName
        query);
    }
    async findAndUpdate(query, value) {
        const req = async () => {
            const result = await this.collection.findOneAndUpdate(query, {
                $set: value,
            }, {
                returnOriginal: true,
            });
            return result.value
                ? { value: result.value, existing: true }
                : { value, existing: false };
        };
        return this.requestWithRetry(req, // request
        "MongoCollection.findAndUpdate", // callerName
        query);
    }
    async updateCore(filter, set, addToSet, upsert) {
        const update = {};
        if (set) {
            update.$set = set;
        }
        if (addToSet) {
            update.$addToSet = addToSet;
        }
        const options = { upsert };
        await this.collection.updateOne(filter, update, options);
    }
    async updateManyCore(filter, set, addToSet, upsert) {
        const update = {};
        if (set) {
            update.$set = set;
        }
        if (addToSet) {
            update.$addToSet = addToSet;
        }
        const options = { upsert };
        await this.collection.updateMany(filter, update, options);
    }
    async requestWithRetry(request, callerName, queryOrFilter) {
        const telemetryProperties = this.getTelemetryPropertiesFromQuery(queryOrFilter);
        return (0, server_services_core_1.requestWithRetry)(request, callerName, telemetryProperties, (e) => this.retryEnabled && this.mongoErrorRetryAnalyzer.shouldRetry(e), // ShouldRetry
        MaxRetryAttempts, // maxRetries
        InitialRetryIntervalInMs, // retryAfterMs
        (error, numRetries, retryAfterInterval) => numRetries * retryAfterInterval, // retryAfterIntervalCalculator
        undefined, /* onErrorFn */ this.telemetryEnabled);
    }
    getTelemetryPropertiesFromQuery(queryOrFilter) {
        const properties = new Map();
        if (!queryOrFilter) {
            return properties;
        }
        if (Object.prototype.hasOwnProperty.call(queryOrFilter, "_id")) {
            properties.set("id", queryOrFilter._id);
        }
        if (Object.prototype.hasOwnProperty.call(queryOrFilter, "tenantId")) {
            properties.set(server_services_telemetry_1.BaseTelemetryProperties.tenantId, queryOrFilter.tenantId);
        }
        if (Object.prototype.hasOwnProperty.call(queryOrFilter, "documentId")) {
            properties.set(server_services_telemetry_1.BaseTelemetryProperties.documentId, queryOrFilter.documentId);
        }
        return properties;
    }
}
exports.MongoCollection = MongoCollection;
class MongoDb {
    constructor(client, retryEnabled = false, telemetryEnabled = false, mongoErrorRetryAnalyzer) {
        this.client = client;
        this.retryEnabled = retryEnabled;
        this.telemetryEnabled = telemetryEnabled;
        this.mongoErrorRetryAnalyzer = mongoErrorRetryAnalyzer;
    }
    // eslint-disable-next-line @typescript-eslint/promise-function-async
    close() {
        return this.client.close();
    }
    on(event, listener) {
        this.client.on(event, listener);
    }
    collection(name) {
        const collection = this.client.db("admin").collection(name);
        return new MongoCollection(collection, this.retryEnabled, this.telemetryEnabled, this.mongoErrorRetryAnalyzer);
    }
    async dropCollection(name) {
        return this.client.db("admin").dropCollection(name);
    }
}
exports.MongoDb = MongoDb;
class MongoDbFactory {
    constructor(config) {
        this.retryEnabled = false;
        this.telemetryEnabled = false;
        const { operationsDbEndpoint, bufferMaxEntries, globalDbEnabled, globalDbEndpoint, connectionPoolMinSize, connectionPoolMaxSize, } = config;
        if (globalDbEnabled) {
            this.globalDbEndpoint = globalDbEndpoint;
        }
        (0, console_1.assert)(!!operationsDbEndpoint, `No endpoint provided`);
        this.operationsDbEndpoint = operationsDbEndpoint;
        this.bufferMaxEntries = bufferMaxEntries;
        this.connectionPoolMinSize = connectionPoolMinSize;
        this.connectionPoolMaxSize = connectionPoolMaxSize;
        this.retryEnabled = config.facadeLevelRetry || false;
        this.telemetryEnabled = config.facadeLevelTelemetry || false;
        this.retryRuleOverride = config.facadeLevelRetryRuleOverride
            ? new Map(Object.entries(config.facadeLevelRetryRuleOverride))
            : new Map();
    }
    async connect(global = false) {
        var _a;
        (0, console_1.assert)(!global || !!this.globalDbEndpoint, `No global endpoint provided
                 when trying to connect to global db.`);
        // Need to cast to any before MongoClientOptions due to missing properties in d.ts
        const options = {
            autoReconnect: true,
            bufferMaxEntries: (_a = this.bufferMaxEntries) !== null && _a !== void 0 ? _a : 50,
            keepAlive: true,
            keepAliveInitialDelay: 180000,
            reconnectInterval: 1000,
            reconnectTries: 100,
            socketTimeoutMS: 120000,
            useNewUrlParser: true,
        };
        if (this.connectionPoolMinSize) {
            options.minSize = this.connectionPoolMinSize;
        }
        if (this.connectionPoolMaxSize) {
            options.poolSize = this.connectionPoolMaxSize;
        }
        const connection = await mongodb_1.MongoClient.connect(global ?
            this.globalDbEndpoint :
            this.operationsDbEndpoint, options);
        const retryAnalyzer = mongoExceptionRetryRules_1.MongoErrorRetryAnalyzer.getInstance(this.retryRuleOverride);
        return new MongoDb(connection, this.retryEnabled, this.telemetryEnabled, retryAnalyzer);
    }
}
exports.MongoDbFactory = MongoDbFactory;
//# sourceMappingURL=mongodb.js.map