/*!
 * Copyright (c) Microsoft Corporation and contributors. All rights reserved.
 * Licensed under the MIT License.
 */
import { IThrottler, IThrottlerHelper, ILogger, IUsageData } from "@fluidframework/server-services-core";
/**
 * A lenient implementation of IThrottlerHelper that prioritizes low latency over strict throttling.
 * This should be used for implementing throttling in places where latency matters more than accuracy,
 * such as service endpoints or socket connections.
 */
export declare class Throttler implements IThrottler {
    private readonly throttlerHelper;
    private readonly minThrottleIntervalInMs;
    private readonly logger?;
    /**
     * Throttling can generate a lot of telemetry, which can be expensive and/or taxing on resources.
     * Use this flag to enable/disable extra telemetry that is useful for validating throttling config correctness.
     * Default: false
     */
    private readonly enableEnhancedTelemetry;
    private readonly lastThrottleUpdateAtMap;
    private readonly countDeltaMap;
    private readonly throttlerResponseCache;
    constructor(throttlerHelper: IThrottlerHelper, minThrottleIntervalInMs?: number, logger?: ILogger, 
    /**
     * Maximum number of keys that should be internally tracked at a given time.
     * Fine tune this and cache age to balance accuracy and memory consumption.
     * If this value is less than number of keys (traffic) per cache age time, the in-memry cache can overflow.
     * Default: 1,000
     */
    maxCacheSize?: number, 
    /**
     * When to mark internal cache values as stale, in milliseconds. In production, this value should not be
     * lower than minThrottleIntervalInMs, otherwise throttle counts will be lost between calculation intervals.
     * Default: 1min
     */
    maxCacheAge?: number, 
    /**
     * Throttling can generate a lot of telemetry, which can be expensive and/or taxing on resources.
     * Use this flag to enable/disable extra telemetry that is useful for validating throttling config correctness.
     * Default: false
     */
    enableEnhancedTelemetry?: boolean);
    /**
     * Increments operation count and calculates throttle status of given operation id.
     * Uses most recently calculated throttle status to determine current throttling, while updating in the background.
     * @throws {@link ThrottlingError} if throttled
     */
    incrementCount(id: string, weight?: number, usageStorageId?: string, usageData?: IUsageData): void;
    /**
     * Decrements operation count of given operation id.
     */
    decrementCount(id: string, weight?: number): void;
    private updateCountDelta;
    private updateAndCacheThrottleStatus;
    private getBaseTelemetryProperties;
}
//# sourceMappingURL=throttler.d.ts.map