"use strict";
/*!
 * Copyright (c) Microsoft Corporation and contributors. All rights reserved.
 * Licensed under the MIT License.
 */
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.Throttler = void 0;
const server_services_core_1 = require("@fluidframework/server-services-core");
const lru_cache_1 = __importDefault(require("lru-cache"));
const server_services_telemetry_1 = require("@fluidframework/server-services-telemetry");
/**
 * A lenient implementation of IThrottlerHelper that prioritizes low latency over strict throttling.
 * This should be used for implementing throttling in places where latency matters more than accuracy,
 * such as service endpoints or socket connections.
 */
class Throttler {
    constructor(throttlerHelper, minThrottleIntervalInMs = 1000000, logger, 
    /**
     * Maximum number of keys that should be internally tracked at a given time.
     * Fine tune this and cache age to balance accuracy and memory consumption.
     * If this value is less than number of keys (traffic) per cache age time, the in-memry cache can overflow.
     * Default: 1,000
     */
    maxCacheSize = 1000, 
    /**
     * When to mark internal cache values as stale, in milliseconds. In production, this value should not be
     * lower than minThrottleIntervalInMs, otherwise throttle counts will be lost between calculation intervals.
     * Default: 1min
     */
    maxCacheAge = 1000 * 60, 
    /**
     * Throttling can generate a lot of telemetry, which can be expensive and/or taxing on resources.
     * Use this flag to enable/disable extra telemetry that is useful for validating throttling config correctness.
     * Default: false
     */
    enableEnhancedTelemetry = false) {
        this.throttlerHelper = throttlerHelper;
        this.minThrottleIntervalInMs = minThrottleIntervalInMs;
        this.logger = logger;
        this.enableEnhancedTelemetry = enableEnhancedTelemetry;
        const cacheOptions = {
            max: maxCacheSize,
            maxAge: maxCacheAge,
        };
        this.lastThrottleUpdateAtMap = new lru_cache_1.default(Object.assign(Object.assign({}, cacheOptions), { dispose: this.enableEnhancedTelemetry ? (key, value) => {
                var _a;
                // Utilize the opportunity to log information before an item is removed from the cache.
                // If a cache entry is removed too soon, it can negatively impact the correctness of throttling.
                const now = Date.now();
                if (now - value < maxCacheAge) {
                    // lastThrottleUpdateAt value should be equal to the time that the cached value was last updated.
                    // If it is being disposed before the maxCacheAge is reached, it indicates that the cache is full.
                    const telemetryProperties = this.getBaseTelemetryProperties(key);
                    const lumberjackProperties = Object.assign(Object.assign({}, telemetryProperties.baseLumberjackProperties), { ageInMs: now - value });
                    (_a = this.logger) === null || _a === void 0 ? void 0 : _a.warn(`Purged lastThrottleUpdateAt for ${key} before maxAge reached`, { messageMetaData: telemetryProperties.baseMessageMetaData });
                    server_services_telemetry_1.Lumberjack.warning(`Purged lastThrottleUpdateAt for ${key} before maxAge reached`, lumberjackProperties);
                }
            } : undefined }));
        this.countDeltaMap = new lru_cache_1.default(cacheOptions);
        this.throttlerResponseCache = new lru_cache_1.default(cacheOptions);
    }
    /**
     * Increments operation count and calculates throttle status of given operation id.
     * Uses most recently calculated throttle status to determine current throttling, while updating in the background.
     * @throws {@link ThrottlingError} if throttled
     */
    incrementCount(id, weight = 1, usageStorageId, usageData) {
        var _a;
        const telemetryProperties = this.getBaseTelemetryProperties(id);
        this.updateCountDelta(id, weight);
        this.updateAndCacheThrottleStatus(id, usageStorageId, usageData)
            .catch((error) => {
            var _a;
            (_a = this.logger) === null || _a === void 0 ? void 0 : _a.error(`Error encountered updating and/or caching throttle status for ${id}: ${error}`, { messageMetaData: telemetryProperties.baseMessageMetaData });
            server_services_telemetry_1.Lumberjack.error(`Error encountered updating and/or caching throttle status for ${id}`, telemetryProperties.baseLumberjackProperties, error);
        });
        // check cached throttle status, but allow operation through if status is not yet cached
        const cachedThrottlerResponse = this.throttlerResponseCache.get(id);
        // eslint-disable-next-line @typescript-eslint/prefer-optional-chain
        if (cachedThrottlerResponse && cachedThrottlerResponse.throttleStatus) {
            const retryAfterInSeconds = Math.ceil(cachedThrottlerResponse.retryAfterInMs / 1000);
            (_a = this.logger) === null || _a === void 0 ? void 0 : _a.info(`Throttled: ${id}`, {
                messageMetaData: Object.assign(Object.assign({}, telemetryProperties.baseMessageMetaData), { reason: cachedThrottlerResponse.throttleReason, retryAfterInSeconds }),
            });
            server_services_telemetry_1.Lumberjack.info(`Throttled: ${id}`, Object.assign(Object.assign({}, telemetryProperties.baseLumberjackProperties), { [server_services_telemetry_1.ThrottlingTelemetryProperties.reason]: cachedThrottlerResponse.throttleReason, [server_services_telemetry_1.ThrottlingTelemetryProperties.retryAfterInSeconds]: retryAfterInSeconds }));
            // eslint-disable-next-line @typescript-eslint/no-throw-literal
            throw new server_services_core_1.ThrottlingError(cachedThrottlerResponse.throttleReason, retryAfterInSeconds);
        }
    }
    /**
     * Decrements operation count of given operation id.
     */
    decrementCount(id, weight = 1) {
        this.updateCountDelta(id, -weight);
    }
    updateCountDelta(id, value) {
        const currentValue = this.countDeltaMap.get(id) || 0;
        this.countDeltaMap.set(id, currentValue + value);
    }
    async updateAndCacheThrottleStatus(id, usageStorageId, usageData) {
        var _a;
        const telemetryProperties = this.getBaseTelemetryProperties(id);
        const now = Date.now();
        if (this.lastThrottleUpdateAtMap.get(id) === undefined) {
            if (this.enableEnhancedTelemetry) {
                (_a = this.logger) === null || _a === void 0 ? void 0 : _a.info(`Starting to track throttling status for ${id}`, { messageMetaData: telemetryProperties.baseMessageMetaData });
                server_services_telemetry_1.Lumberjack.info(`Starting to track throttling status for ${id}`, telemetryProperties.baseLumberjackProperties);
            }
            this.lastThrottleUpdateAtMap.set(id, now);
        }
        const lastThrottleUpdateTime = this.lastThrottleUpdateAtMap.get(id);
        if (now - lastThrottleUpdateTime > this.minThrottleIntervalInMs) {
            const countDelta = this.countDeltaMap.get(id);
            this.lastThrottleUpdateAtMap.set(id, now);
            this.countDeltaMap.set(id, 0);
            const messageMetaData = Object.assign(Object.assign({}, telemetryProperties.baseMessageMetaData), { weight: countDelta });
            const lumberjackProperties = Object.assign(Object.assign({}, telemetryProperties.baseLumberjackProperties), { [server_services_telemetry_1.ThrottlingTelemetryProperties.weight]: countDelta });
            // poplulate usageData with relevant data.
            if (usageData) {
                usageData.value = countDelta;
                usageData.startTime = lastThrottleUpdateTime;
                usageData.endTime = now;
            }
            await this.throttlerHelper.updateCount(id, countDelta, usageStorageId, usageData)
                .then((throttlerResponse) => {
                var _a;
                if (this.enableEnhancedTelemetry) {
                    (_a = this.logger) === null || _a === void 0 ? void 0 : _a.info(`Incremented throttle count for ${id} by ${countDelta}`, { messageMetaData });
                    server_services_telemetry_1.Lumberjack.info(`Incremented throttle count for ${id} by ${countDelta}`, lumberjackProperties);
                }
                this.throttlerResponseCache.set(id, throttlerResponse);
            })
                .catch((err) => {
                var _a;
                (_a = this.logger) === null || _a === void 0 ? void 0 : _a.error(`Failed to update throttling count for ${id}: ${err}`, { messageMetaData });
                server_services_telemetry_1.Lumberjack.error(`Failed to update throttling count for ${id}`, lumberjackProperties, err);
            });
        }
    }
    getBaseTelemetryProperties(key) {
        return {
            baseMessageMetaData: {
                key,
                eventName: "throttling",
            },
            baseLumberjackProperties: {
                [server_services_telemetry_1.CommonProperties.telemetryGroupName]: "throttling",
                [server_services_telemetry_1.ThrottlingTelemetryProperties.key]: key,
            },
        };
    }
}
exports.Throttler = Throttler;
//# sourceMappingURL=throttler.js.map