"use strict";
/*!
 * Copyright (c) Microsoft Corporation and contributors. All rights reserved.
 * Licensed under the MIT License.
 */
Object.defineProperty(exports, "__esModule", { value: true });
exports.ThrottlerHelper = void 0;
/**
 * Implements the Token Bucket algorithm to calculate rate-limiting for throttling operations.
 */
class ThrottlerHelper {
    constructor(throttleAndUsageStorageManager, rateInOperationsPerMs = 1000000, operationBurstLimit = 1000000, minCooldownIntervalInMs = 1000000) {
        this.throttleAndUsageStorageManager = throttleAndUsageStorageManager;
        this.rateInOperationsPerMs = rateInOperationsPerMs;
        this.operationBurstLimit = operationBurstLimit;
        this.minCooldownIntervalInMs = minCooldownIntervalInMs;
    }
    async updateCount(id, count, usageStorageId, usageData) {
        const now = Date.now();
        let throttlingMetric = await this.throttleAndUsageStorageManager.getThrottlingMetric(id);
        if (!throttlingMetric) {
            // start a throttling metric with 1 operation burst limit's worth of tokens
            throttlingMetric = {
                count: this.operationBurstLimit,
                lastCoolDownAt: now,
                throttleStatus: false,
                throttleReason: undefined,
                retryAfterInMs: 0,
            };
        }
        // Exit early if already throttled and no chance of being unthrottled
        const retryAfterInMs = this.getRetryAfterInMs(throttlingMetric, now);
        if (retryAfterInMs > 0) {
            throttlingMetric.retryAfterInMs = retryAfterInMs;
            await this.setThrottlingMetricAndUsageData(id, throttlingMetric, usageStorageId, usageData);
            return this.getThrottlerResponseFromThrottlingMetrics(throttlingMetric);
        }
        // replenish "tokens" if possible
        const amountToReplenish = this.getTokenReplenishAmount(throttlingMetric, now);
        if (amountToReplenish > 0) {
            throttlingMetric.count += amountToReplenish;
            throttlingMetric.lastCoolDownAt = now;
        }
        // adjust "tokens" based on given count
        throttlingMetric.count -= count;
        // throttle if "token bucket" is empty
        const newRetryAfterInMs = this.getRetryAfterInMs(throttlingMetric, now);
        if (newRetryAfterInMs > 0) {
            throttlingMetric.throttleStatus = true;
            throttlingMetric.throttleReason =
                `Throttling count exceeded by ${Math.abs(throttlingMetric.count)} at ${new Date(now).toISOString()}`;
            throttlingMetric.retryAfterInMs = newRetryAfterInMs;
        }
        else {
            throttlingMetric.throttleStatus = false;
            throttlingMetric.throttleReason = "";
            throttlingMetric.retryAfterInMs = 0;
        }
        await this.setThrottlingMetricAndUsageData(id, throttlingMetric, usageStorageId, usageData);
        return this.getThrottlerResponseFromThrottlingMetrics(throttlingMetric);
    }
    async getThrottleStatus(id) {
        const throttlingMetric = await this.throttleAndUsageStorageManager.getThrottlingMetric(id);
        if (!throttlingMetric) {
            return undefined;
        }
        return this.getThrottlerResponseFromThrottlingMetrics(throttlingMetric);
    }
    async setThrottlingMetricAndUsageData(id, throttlingMetric, usageStorageId, usageData) {
        await (usageStorageId && usageData ? this.throttleAndUsageStorageManager.setThrottlingMetricAndUsageData(id, throttlingMetric, usageStorageId, usageData) : this.throttleAndUsageStorageManager.setThrottlingMetric(id, throttlingMetric));
    }
    getThrottlerResponseFromThrottlingMetrics(throttlingMetric) {
        return {
            throttleStatus: throttlingMetric.throttleStatus,
            throttleReason: throttlingMetric.throttleReason,
            retryAfterInMs: throttlingMetric.retryAfterInMs,
        };
    }
    getTokenReplenishAmount(throttlingMetric, now) {
        const timeSinceLastCooldownInMs = now - throttlingMetric.lastCoolDownAt;
        // replenish "tokens" at most once per minCooldownInterval
        if (timeSinceLastCooldownInMs > this.minCooldownIntervalInMs) {
            const tokensToReplenish = Math.floor(timeSinceLastCooldownInMs * this.rateInOperationsPerMs);
            // don't let the bucket overflow
            if (tokensToReplenish + throttlingMetric.count > this.operationBurstLimit) {
                return this.operationBurstLimit - throttlingMetric.count;
            }
            return tokensToReplenish;
        }
        return 0;
    }
    getRetryAfterInMs(throttlingMetric, now) {
        const tokenDebt = 0 - throttlingMetric.count;
        const amountPossibleToReplenishNow = this.getTokenReplenishAmount(throttlingMetric, now);
        const timeUntilNextCooldown = throttlingMetric.lastCoolDownAt + this.minCooldownIntervalInMs - now;
        const remainingTokenDebt = tokenDebt - amountPossibleToReplenishNow;
        const timeUntilDebtReplenished = remainingTokenDebt / this.rateInOperationsPerMs;
        if (timeUntilDebtReplenished <= 0) {
            // no need to wait because tokens can be replenished to satisfactory amount
            return timeUntilDebtReplenished;
        }
        // must at least wait until next cooldown
        return Math.max(timeUntilNextCooldown, timeUntilDebtReplenished);
    }
}
exports.ThrottlerHelper = ThrottlerHelper;
//# sourceMappingURL=throttlerHelper.js.map